// Copyright 2009-2021 Intel Corporation
// SPDX-License-Identifier: Apache-2.0

#include "parallel_any_of.cpp"
#include "parallel_filter.cpp"
#include "parallel_for.cpp"
#include "parallel_for_for.cpp"
#include "parallel_for_for_prefix_sum.cpp"
#include "parallel_map.cpp"
#include "parallel_partition.cpp"
#include "parallel_prefix_sum.cpp"
#include "parallel_reduce.cpp"
#include "parallel_set.cpp"
#include "parallel_sort.cpp"
