// Copyright 2009-2021 Intel Corporation
// SPDX-License-Identifier: Apache-2.0

#include "../include/embree4/rtcore_config.h"

1 VERSIONINFO
 FILEVERSION RTC_VERSION_MAJOR,RTC_VERSION_MINOR,RTC_VERSION_PATCH,0
 PRODUCTVERSION RTC_VERSION_MAJOR,RTC_VERSION_MINOR,RTC_VERSION_PATCH,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "Intel"
            VALUE "FileDescription", "Embree Ray Tracing Kernels"
            VALUE "FileVersion", RTC_VERSION_STRING
            VALUE "ProductVersion", RTC_VERSION_STRING
            VALUE "LegalCopyright", "(C) 2009-2021 Intel Corporation"
            VALUE "InternalName", "Embree"
            VALUE "ProductName", "Intel(R) Embree"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
