// Copyright 2009-2021 Intel Corporation
// SPDX-License-Identifier: Apache-2.0

#include "../../include/embree4/rtcore.h"
#include "../../include/embree4/rtcore_builder.h"
#include "../../include/embree4/rtcore_ray.h"

// this C-file just tests if our API headers are C99 compliant
