// Copyright 2009-2021 Intel Corporation
// SPDX-License-Identifier: Apache-2.0

// #include "../../../external/catch.hpp"
// #include "../../../../kernels/common/scene_line_segments.h"

// using namespace embree;

// TEST_CASE ("Segment Left/Right exists tests - Issue 726", "[scene_line_segments]") 
// {
//   // embree::Device device ("");
//   // LineSegments lines (&device, Geometry::GTY_ROUND_LINEAR_CURVE);
// }
